"""
Author: Juha Meskanen
Date: 2024-10-26
"""

from typing import Any, Optional
from typing_extensions import override
import unittest
from unittest.mock import patch, MagicMock
from masterpiece.plugmaster import PlugMaster
from masterpiece.plugin import Plugin
from masterpiece.composite import Composite
from masterpiece.masterpiece import MasterPiece


class TestPluginFoo(Plugin):
    """Mock plugin class to simulate plugin behavior."""

    def __init__(self, name: str) -> None:
        super().__init__(name)
        # self.install = MagicMock()  # Mock the install method

    @override
    def install(self, app: Composite) -> None:
        app.add(self)
        print(f"plugin {self.name} installed to {app}")


class TestPlugMaster(unittest.TestCase):

    def setUp(self) -> None:
        self.app_name = "test_app"
        self.plugmaster = PlugMaster(self.app_name)
        self.app = Composite(self.app_name)
        self.app.add = MagicMock()  # Mock the add method

        # Define the TestPlugin class
        class Test2Plugin(Plugin):

            def install(self, app: Composite) -> None:
                app.add(self)  # Simulate adding the plugin to the app

        # Add the class to the plugmaster's plugins
        Test2Plugin.__name__ = "Test2Plugin"  # Define __name__ attribute

        self.plugmaster.plugins["TestPlugin"] = Test2Plugin

    @patch("importlib.metadata.entry_points")
    def test_load_plugins(self, mock_entry_points: Any) -> None:
        # Create a mock plugin class
        mock_plugin_class = MagicMock()
        mock_plugin_class.__name__ = "Test2Plugin"  # Define __name__ attribute
        mock_plugin_instance = mock_plugin_class.return_value  # Instance of the plugin
        mock_plugin_instance.install = MagicMock()  # Mock the install method

        # Create a mock entry point
        mock_entry_point = MagicMock()
        mock_entry_point.load.return_value = mock_plugin_class

        # Ensure that entry_points returns a list with our mock entry point
        mock_entry_points.return_value.select.return_value = [mock_entry_point]

        # Call the load method
        self.plugmaster.load("test_app.plugins")

        # Verify that the plugin was added
        self.assertIn("TestPlugin", self.plugmaster.plugins)

    def test_find_class_by_name(self) -> None:
        mock_class = MagicMock(spec=MasterPiece)
        self.plugmaster.plugins["TestPlugin"] = mock_class

        found_class = self.plugmaster.find_class_by_name("TestPlugin")
        self.assertEqual(found_class, mock_class)

        not_found_class = self.plugmaster.find_class_by_name("NonExistentPlugin")
        self.assertIsNone(not_found_class)

    def test_instantiate_class_by_name(self) -> None:
        obj = self.plugmaster.instantiate_class_by_name(self.app, "TestPlugin")
        self.assertIsNotNone(obj)

        # Ensure the add method is called
        self.app.add.assert_called()

    def test_instantiate_class_by_name_with_plugin(self) -> None:
        class MockPlugin(Plugin):
            @override
            def install(self, app: Composite) -> None:
                app.add(self)  # Simulate adding the plugin to the app

        # Add MockPlugin class to the plugmaster
        self.plugmaster.plugins["TestPlugin"] = MockPlugin

        obj: Optional[MasterPiece] = self.plugmaster.instantiate_class_by_name(
            self.app, "TestPlugin"
        )
        self.assertIsNotNone(obj)

        # Verify that install is called on the plugin instance
        obj.install(self.app)  # Simulate calling install
        self.assertTrue(self.app.add.called)

    def test_install_plugins(self) -> None:
        # Call the install method, which should invoke the install method of the plugin class
        self.plugmaster.install(self.app)
        self.assertTrue(self.app.add.called)  # Ensure add was called

    def test_get_plugins(self) -> None:
        mock_class = MagicMock(spec=MasterPiece)
        self.plugmaster.plugins["TestPlugin"] = mock_class

        plugins = self.plugmaster.get()
        self.assertIn("TestPlugin", plugins)
        self.assertEqual(plugins["TestPlugin"], mock_class)

    def test_plugin_install(self) -> None:
        mock_app = MagicMock(spec=Composite)
        plugin = TestPluginFoo("test plugin")

        # Call the install method
        plugin.install(mock_app)

        # Verify the install method was called with the mock app
        self.assertTrue(mock_app.add.called)  # Ensure add was called


if __name__ == "__main__":
    unittest.main()



"""
This module exposes the core concrete classes of the framework, which are intended to be 
instantiated and used by application developers. 

The core package consists of two types of classes:

1. **Abstract Base Classes (ABCs)**: These define the interfaces that framework implementors should 
    follow when creating subclasses. They provide method contracts that must be implemented by concrete 
    subclasses. These classes are **not intended to be instantiated** directly.

2. **Concrete Classes**: These are the classes that **can be instantiated** and used directly by application 
    developers. They represent fully implemented functionality and are ready for use in applications.

This structure ensures clear separation between the framework's core functionality and the classes that can be
used to build applications on top of the framework.

"""

from .masterpiece import MasterPiece, classproperty
from .composite import Composite
from .application import Application
from .log import Log
from .plugin import Plugin
from .plugmaster import PlugMaster
from .treevisualizer import TreeVisualizer
from .url import URL
from .format import Format
from .jsonformat import JsonFormat
from .masterpiecethread import MasterPieceThread
from .timeseries import TimeSeries, Measurement
from .mqtt import Mqtt, MqttMsg

__all__ = [
    "MasterPiece",
    "Composite",
    "Application",
    "Log",
    "Plugin",
    "PlugMaster",
    "ArgsMaestro",
    "TreeVisualizer",
    "classproperty",
    "URL",
    "Format",
    "JsonFormat",
    "MasterPieceThread",
    "TimeSeries",
    "Mqtt",
    "MqttMsg",
    "Measurement",
]



from typing import Any, Optional, Union
from typing_extensions import override
from .masterpiece import MasterPiece

from abc import ABC, abstractmethod
from typing import Dict, Any


class Measurement(ABC):
    """Abstract base class for measurement structures."""

    def __init(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the measurement into a dictionary format.
        Used for writing to storage or transferring data.
        """
        pass

    @abstractmethod
    def from_dict(self, data: Dict[str, Any]) -> "Measurement":
        """
        Populate the measurement fields from a dictionary.
        """
        pass

    @abstractmethod
    def validate(self) -> bool:
        """
        Validate the measurement data.
        Return True if valid, otherwise raise an exception or return False.
        """
        pass

    @abstractmethod
    def tag(self, tag: str, value: str) -> "Measurement":
        """
        Add a tag to the measurement and return self for method chaining.
        """
        raise Exception("tag not implemented")

    @abstractmethod
    def field(self, field: str, value: Union[int, float, str, bool]) -> "Measurement":
        """
        Add a field to the measurement and return self for method chaining.
        """
        raise Exception("field not implemented")

    @abstractmethod
    def time(self, timestamp: Union[str, int]) -> "Measurement":
        """
        Set the timestamp for the measurement and return self for method chaining.
        """
        raise Exception("time not implemented")


class TimeSeries(MasterPiece):
    """An abstract base class for time series database interactions.

    This class defines a standardized interface for reading and writing time series data.
    It serves as a foundation for implementing support for specific time series databases,
    abstracting low-level details and enabling consistent access patterns.

    Subclasses must implement the `write_dict` and `read_dict` methods to provide
    database-specific functionality.

    """

    token: str = ""
    org: str = ""
    host: str = ""
    database: str = ""

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def write(self, point: Measurement) -> None:
        """Write record to database table.

        Args:
            point (Any): The data point to be written to the database.

        Raises:
            Exception: If not implemented in the subclass.
        """
        raise Exception("write not implemented")

    def write_dict(
        self, name: str, tags: dict[str, Any], fields: dict[str, Any], ts: str
    ) -> None:
        """Write record to the database table.

        Args:
            name (str): The name of the measurement.
            tags (dict[str, Any]): Tags (indexed keys) for filtering the data.
            fields (dict[str, Any]): Measurement data fields.
            ts (str): The timestamp for the measurement.

        Returns:
            None

        Raises:
            Exception: If the write method is not implemented in the subclass.
        """
        raise Exception("TimeSeries write_dict not implemented")

    def read_dict(
        self,
        measurement: str,
        start_time: str,
        end_time: Optional[str] = None,
        tags: Optional[dict[str, Any]] = None,
        fields: Optional[list[str]] = None,
    ) -> list[dict[str, Any]]:
        """Reads records from the database using SQL.

        Args:
            measurement (str): The name of the measurement (table) to query.
            start_time (str): The start time for the query (ISO8601 format).
            end_time (Optional[str]): The end time for the query (ISO8601 format). Defaults to None.
            tags (Optional[dict[str, Any]]): Tags to filter the data (as WHERE conditions). Defaults to None.
            fields (Optional[list[str]]): Specific fields to include in the result. Defaults to None.

        Returns:
            list[dict[str, Any]]: A list of records matching the query.

        Raises:
            Exception: If reading from the database fails.
        """
        raise Exception("TimeSeries read_dict not implemented")

    def read_last_value(
        self,
        measurement: str,
        tags: Optional[dict[str, Any]] = None,
        fields: Optional[list[str]] = None,
    ) -> dict[str, Any]:
        """Reads the last value from the database.

        Args:
            measurement (str): The name of the measurement (table) to query.
            tags (Optional[dict[str, Any]]): Tags to filter the data (as WHERE conditions). Defaults to None.
            fields (Optional[list[str]]): Specific fields to include in the result. Defaults to None.

        Returns:
            list[dict[str, Any]]: A list of records matching the query.

        Raises:
            Exception: If reading from the database fails.
        """
        raise Exception("TimeSeries read_last_value not implemented")

    def read_point(
        self,
        measurement: str,
        start_time: str,
        end_time: Optional[str] = None,
        tags: Optional[dict[str, Any]] = None,
        fields: Optional[list[str]] = None,
    ) -> list[Measurement]:
        """Reads records from the database and returns them as Point objects.

        Args:
            measurement (str): The name of the measurement (table) to query.
            start_time (str): The start time for the query (ISO8601 format).
            end_time (Optional[str]): The end time for the query (ISO8601 format). Defaults to None.
            tags (Optional[Dict[str, Any]]): Tags to filter the data (as WHERE conditions). Defaults to None.
            fields (Optional[list[str]]): Specific fields to include in the result. Defaults to None.

        Returns:
            List[Point]: A list of Point objects matching the query.

        Raises:
            Exception: If reading from the database fails.
        """
        raise Exception("TimeSeries read_point not implemented")

    @abstractmethod
    def measurement(self, measurement: str) -> Measurement:
        """Create new measurement.

        Args:
            measurement (str): The name of the measurement (table) to query.

        Returns:
            Measurement

        Raises:
            NotImplementedError: If this method is not implemented by the subclass.
        """
        raise NotImplementedError("Measurement not implemented")

    @override
    def to_dict(self) -> dict[str, Any]:
        data: dict[str, Any] = super().to_dict()
        data["_timeseries"] = {}
        attributes = ["host", "org", "database", "token"]
        for attr in attributes:
            if getattr(self, attr) != getattr(type(self), attr):
                data["_timeseries"][attr] = getattr(self, attr)
        return data

    @override
    def from_dict(self, data_dict: dict[str, Any]) -> None:
        super().from_dict(data_dict)
        for key, value in data_dict["_timeseries"].items():
            setattr(self, key, value)



"""
The elementary base class of everything.

Author: Juha Meskanen
Date: 2024-10-26
"""

from __future__ import annotations

import inspect
import logging
from typing import Any, Callable, Optional, Type, Union, cast
from .url import URL


class classproperty:
    """
    A decorator that allows you to define class-level properties.
    Replaces the deprecated combination of @classmethod and @property.
    """

    def __init__(self, func: Callable[[Type[Any]], Any]) -> None:
        """
        Initialize the classproperty with the function to decorate.

        Args:
            func: A callable that accepts a class type and returns any value.
        """
        self.func = func

    def __get__(self, instance: Any, owner: Type[Any]) -> Any:
        """
        Get the class property value by calling the decorated function with the class.

        Args:
            instance: The instance of the class (not used here).
            owner: The class that owns this property.

        Returns:
            The result of calling the decorated function with the class.
        """
        return self.func(owner)


class MasterPiece:
    """An object with a name. Base class of everything. Serves as the
    foundational base class for any real-world object that can be a part of
    a hierarchy.

    """

    # non-serializable private class attributes
    _log: Optional[logging.Logger] = None
    _factory: dict[str, Type[MasterPiece]] = {}

    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        """Called when a new sub-class is created.

        Automatically registers the sub class by calling its register()
        method. For more information on this method consult Python
        documentation.
        """
        super().__init_subclass__(**kwargs)
        cls.register()

    @classmethod
    def classattrs_to_dict(cls) -> dict[str, Any]:
        """Convert the class's own attributes to a dictionary, excluding inherited and private ones."""
        return {
            attr: value
            for attr, value in cls.__dict__.items()
            if not callable(value)  # Exclude regular methods
            and not isinstance(
                value, (classmethod, staticmethod, classproperty)
            )  # Exclude class/staticmethods
            and not attr.startswith("__")  # Exclude special attributes like __module__
            and not attr.startswith("_")  # Exclude private/protected attributes
        }

    @classmethod
    def classattrs_from_dict(cls, attributes: dict[str, Any]) -> None:
        """Set only the class's own attributes from a dictionary."""
        # Get the set of attributes defined in this class itself
        own_class_attrs = set(cls.__dict__.keys())

        for key, value in attributes.items():
            # Only set attributes that belong directly to this class (not inherited)
            if key in own_class_attrs:
                setattr(cls, key, value)

    @classmethod
    def has_class_method_directly(cls, method_name: str) -> bool:
        """
        Check if the method is defined directly in the class (not inherited).
        """
        method = getattr(cls, method_name, None)
        # Ensure that the method is callable and check its defining class
        if callable(method):
            # Check if the method's class is the same as the one we are querying
            return cast(bool, method.__qualname__.split(".")[0] == cls.__name__)
        return False

    @classmethod
    def factory(cls) -> dict[str, Type[MasterPiece]]:
        """Fetch the dictionary holding class names and associated classes.

        Returns:
            factory: with class names and associated classes
        """
        return cls._factory

    @classmethod
    def register(cls) -> None:
        """Register the class.

        Called immediately upon class initialization, right before the class attributes
        are loaded from the class specific configuration files.

        Subclasses can extend this with custom register functionality:

        .. code-block:: python

            class MyMasterPiece(MasterPiece):

                @classmethod
                def register(cls):
                    super().register()  # Don't forget
                    cls._custom_field = True
        """
        cls.init_class(cls)

    @classmethod
    def init_class(cls, clazz: Type[MasterPiece]) -> None:
        """Initialize class.  Registers the class into the
        class factory .

        Args:
            clazz (class): class to be initialized
        """
        if clazz.__name__ not in cls._factory:
            if inspect.isabstract(clazz):
                cls.log_info(f"Abstract {clazz.__name__} skipped")
            else:
                cls._factory[clazz.__name__] = clazz
                cls.log_info(f"Class {clazz.__name__} initialized")

    @classmethod
    def set_log(cls, l: logging.Logger) -> None:
        """Set logger.

        Args:
            l (logger): logger object
        """

        cls._log = l

    @classmethod
    def get_class_id(cls) -> str:
        """Return the class id of the class. Each class has an unique
        name that can be used for instantiating the class via
        :meth:`Object.instantiate` method.

        Args:
            cls (class): class

        Returns:
            id (str): unique class identifier through which the class can be
            instantiated by factory method pattern.
        """
        return cls.__name__

    def __init__(
        self,
        name: str = "noname",
        payload: Optional[MasterPiece] = None,
        parent: Optional[MasterPiece] = None,
    ):
        """
        Initialize a MasterPiece object.

        Args:
            name (str): The name of the object.
            payload (MasterPiece, optional): Optional payload data associated with the object. Defaults to None.
            parent (MasterPiece, optional): The parent object in the hierarchy. Defaults to None.
        """
        self.name = name
        self.payload = payload
        self.parent = parent
        self._elapsed: float = 0.0
        self._num_updates: int = 0

    @classmethod
    def log_debug(cls, msg: str, details: str = "") -> None:
        """Logs the given debug message to the application log.

        Args:
            msg (str): The  message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        full_message = f"{cls.__name__} : {msg}"
        if details:
            full_message += f" - {details}"
        if cls._log is not None:
            cls._log.debug(full_message)
        else:
            print(full_message)

    @classmethod
    def log_warning(cls, msg: str, details: str = "") -> None:
        """Logs the given debug message to the application log.

        Args:
            msg (str): The  message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        full_message = f"{cls.__name__} : {msg}"
        if details:
            full_message += f" - {details}"
        if cls._log is not None:
            cls._log.warning(full_message)
        else:
            print(full_message)

    @classmethod
    def log_info(cls, msg: str, details: str = "") -> None:
        """Logs the given  message to the application log.

        Args:
            msg (str): The  message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        full_message = f"{cls.__name__} : {msg}"
        if details:
            full_message += f" - {details}"
        if cls._log is not None:
            cls._log.info(full_message)
        else:
            print(full_message)

    @classmethod
    def log_error(cls, msg: str, details: str = "") -> None:
        """Logs the given  message to the application log.

        Args:
            msg (str): The  message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        full_message = f"{cls.__name__} : {msg}"
        if details:
            full_message += f" - {details}"
        if cls._log is not None:
            cls._log.error(full_message)
        else:
            print(full_message)

    def debug(self, msg: str, details: str = "") -> None:
        """Logs the given debug message to the application log.

        Args:
            msg (str): The information message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        self.log_debug(f"{self.name} : {msg}", details)

    def info(self, msg: str, details: str = "") -> None:
        """Logs the given information message to the application log.

        Args:
            msg (str): The information message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        self.log_info(f"{self.name} : {msg}", details)

    def warning(self, msg: str, details: str = "") -> None:
        """Logs the given warning message to the application log.

        Args:
            msg (str): The message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        self.log_warning(f"{self.name} : {msg}", details)

    def error(self, msg: str, details: str = "") -> None:
        """Logs the given error message to the application log.

        Args:
            msg (str): The message to be logged.
            details (str): Additional detailed information for the message to be logged
        """
        self.log_error(f"{self.name} : {msg}", details)

    def to_dict(self) -> dict[str, Any]:
        """Convert instance attributes to a dictionary."""

        return {
            "_class": self.get_class_id(),  # the real class
            "_version": 0,
            "_object": {
                "name": self.name,
                "payload": (
                    self.payload.to_dict() if self.payload is not None else None
                ),
            },
        }

    def from_dict(self, data: dict[str, Any]) -> None:
        """Update instance attributes from a dictionary."""

        if self.get_class_id() != data["_class"]:
            raise ValueError(
                f"Class mismatch, expected:{self.get_class_id()}, actual:{data['_class']}"
            )
        for key, value in data["_object"].items():
            if key == "payload":
                if value is not None:
                    self.payload = MasterPiece.instantiate(value["_class"])
                    self.payload.from_dict(value)
                else:
                    self.payload = None
            else:
                setattr(self, key, value)

    def copy(self) -> MasterPiece:
        """Create and return a copy of the current object.

        This method serializes the current object to a dictionary using the `to_dict` method,
        creates a new instance of the object's class, and populates it with the serialized data
        using the `from_dict` method.

        This method uses class identifier based instantiation (see factory method pattern) to
        create a new instance of the object, and 'to_dict' and 'from_dict'  methods to initialize
        object's state.

        Returns:
            A new instance of the object's class with the same state as the original object.

        Example:
        ::

            clone_of_john = john.copy()
        """

        data = self.to_dict()
        copy_of_self = MasterPiece.instantiate(self.get_class_id())
        copy_of_self.from_dict(data)
        return copy_of_self

    def do(
        self,
        action: Callable[["MasterPiece", dict[str, Any]], bool],
        context: dict[str, Any],
    ) -> bool:
        """
        Execute the given action to the object, by calling the provided `action`.

        Args:
            action(Callable[["MasterPiece", dict[str, Any]], bool]): A callable that takes
            (node, context) and returns a boolean.
            context (dict[str, Any]): Any context data that the action may use.

        Returns:
            The return value from the executed action.
        """
        return action(self, context)

    def run(self) -> None:
        """Run the masterpiece.  Dispatches the call to `payload` object and
        returns  the control to the caller.
        """
        if self.payload is not None and isinstance(self.payload, MasterPiece):
            self.payload.run()

    def run_forever(self) -> None:
        """Run the payload forever. This method will return only when violently
        terminated. If the object does not have playload object, or it is not
        instance of 'MasterPiece' class then returns immediately and this method has
        no effect.
        """
        if self.payload is not None and isinstance(self.payload, MasterPiece):
            try:
                self.payload.run_forever()
                print("Newtorking loop exit without exception")
            except KeyboardInterrupt:
                print("Application interrupted by user.")
            except (ValueError, IOError) as e:
                print(f"Specific error occurred: {e}")
            except Exception as e:
                print(f"An error occurred: {e}")

    def shutdown(self) -> None:
        """Shutdown the payload object. If the payload object is None, or is not instance of MasterPiece,
        then the call has no effect.
        """
        if self.payload is not None and isinstance(self.payload, MasterPiece):
            self.payload.shutdown()

    @classmethod
    def instantiate(cls, class_id: str, *args: Any) -> MasterPiece:
        """Create an instance of the class corresponding to the given class identifier.

        Args:
            class_id (str): Identifier of the class to instantiate.
            *args: Optional arguments to pass to the class constructor.

        Returns:
            MasterPiece: An instance of the class corresponding to the given class identifier.
        """
        if class_id in cls._factory:
            return cls._factory[class_id](*args)
        raise ValueError(f"Attempting to instantiate unregistered class {class_id}")

    @classmethod
    def find_class(cls, class_id: str) -> Union[Type[MasterPiece], None]:
        """Create an instance of the class corresponding to the given class identifier.

        Args:
            class_id (str): Identifier of the class to instantiate.
            *args: Optional arguments to pass to the class constructor.

        Returns:
            MasterPiece: An instance of the class corresponding to the given class identifier.
        """
        if class_id in cls._factory:
            return cls._factory[class_id]
        return None

    @property
    def parent(self) -> Optional[MasterPiece]:
        return self._parent

    @parent.setter
    def parent(self, value: Optional[MasterPiece]) -> None:
        self._parent = value

    def make_url(self) -> URL:
        """Generate the URL for the composite, including all children."""
        if self.parent:
            parent_url = self.parent.make_url()
            return URL(f"{parent_url}/{self.name}")
        return URL(f"/{self.name}")

    def root(self) -> MasterPiece:
        """Fetch the root object

        Returns:
            MasterPiece: root object
        """
        if self.parent:
            return self.parent.root()
        return self

    def resolve_url(self, url: URL) -> Optional["MasterPiece"]:
        """Find a MasterPiece in the hierarchy matching the URL."""
        segments = url.segments

        if len(segments) == 0:
            return self  # Base case: if URL is empty, return self (root)

        if segments[0] == ".":
            # Current object, continue search with the rest of the segments
            return self.resolve_url(URL("/".join(segments[1:])))

        if segments[0] == "..":
            # Parent object, move to the parent (if exists)
            if self.parent:
                # Remove the ".." segment before passing the rest to the parent
                return self.parent.resolve_url(URL("/".join(segments[1:])))
            return None  # No parent to traverse up

        # Regular segment comparison (matching this object's name)
        if segments[0] == self.name:
            if len(segments) == 1:
                return self  # Found a match for the current segment
            # Continue search in parent, remove the first segment from the URL
            if self.parent:
                return self.parent.resolve_url(URL("/".join(segments[1:])))

        return None  # If nothing matches, return None

    def update_metrics(self, elapsed: float) -> None:
        """Updates the number of `update()` methods calld, and time spent in the method.
        This method is called internally from the `update()` method.

        Args:
            elapsed (float): elapsed seconds.
        """
        self._elapsed += elapsed
        self._num_updates += 1

    def acquire_time_spent(self) -> float:
        """Reads off the average time the thread has spent in its `update()` method since
        the last call, and resets the statistics for the next measurement cycle. This method can be used
        for monitoring healtiness of the thread.

        Returns:
            elapsed (float): elapsed seconds.
        """
        if self._num_updates > 0:
            time_spent: float = self._elapsed * 1000.0 / self._num_updates
            self._num_updates = 0
            self._elapsed = 0.0
            return time_spent
        else:
            return 0.0


# Register MasterPiece manually since __init_subclass__() won't be called on it.
MasterPiece.register()



"""
Author: Juha Meskanen
Date: 2024-10-26
"""

from io import IOBase
import json
from typing import Type
from typing_extensions import override

from .masterpiece import MasterPiece, classproperty
from .format import Format


class JsonFormat(Format):
    """
    The `JsonFormat` class provides methods for serializing and deserializing objects
    to and from JSON format.

    Features:
    ---------
    - Serializes object attributes to a JSON file or stream.
    - Deserializes object attributes from a JSON file or stream.

    Usage:
    ------
    To use the `JsonFormat`, create an instance by passing the target stream. Then,
    call the `serialize` or `deserialize` method with the appropriate object.

    Example:
    --------
    .. code-block:: python

        from masterpiece.core import JsonFormat, MasterPiece

        # Create a JsonFormat instance with a file stream
        with open("output.json", "w") as f:
            json_format = JsonFormat(f)
            json_format.serialize(piece)  # piece is the object to serialize

        with open("output.json", "r") as f:
            json_format = JsonFormat(f)
            json_format.deserialize(piece)  # piece is the object to deserialize
    """

    @override
    def __init__(self, stream: IOBase) -> None:
        """Initialize the JsonFormat with a stream (file object).

        Args:
            stream (Any): The stream to write/read JSON data.
        """
        super().__init__(stream)

    @override
    def serialize(self, obj: MasterPiece) -> None:
        """Serialize the object to the given JSON stream.

        Args:
            obj (Any): The object to serialize.
        """
        json.dump(obj.to_dict(), self.stream, indent=4)

    @override
    def deserialize(self, obj: MasterPiece) -> None:
        """Load attributes from the given JSON stream into the object.

        Args:
            obj (Any): The object to deserialize into.
        """
        attributes = json.load(self.stream)
        obj.from_dict(attributes)

    @override
    def save_configuration(self, clazz: Type[MasterPiece]) -> None:
        """Create class configuration file, if configuration is enabled and
        if the file does not exist yet. See --config startup argument.
        Args:
            clazz (Type[Piece]) class to be saved

        """
        json.dump(clazz.classattrs_to_dict(), self.stream)

    @override
    def load_configuration(self, clazz: Type[MasterPiece]) -> None:
        """Load class attributes from a JSON file.
        Args:
            clazz (Type[Piece]) class to be configured
        """
        clazz.classattrs_from_dict(json.load(self.stream))

    # TODO: Fix this, if you can, beats me. According to mypy:
    # Argument 1 to "classproperty" has incompatible type "Callable[[JsonFormat], str]"
    # expected "Callable[[type[Any]], Any]"Mypy
    @classproperty  # type: ignore
    def file_extension(cls) -> str:
        """Fetch the file extension

        Returns:
            str: file extension including the dot
        """
        return ".json"



"""
Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import sys
import os

# pull in the default Sphinx stuff from masterpiece framework
sys.path.insert(0, os.path.join("..", "..", "config"))
from sphinxconf import *

# -- Project specific info ---
project: str = "masterpiece"
copyright: str = "2024, juha meskanen"
author: str = "juha meskanen"



"""
Path for identifying objects in a hierarchy.

Author: Juha Meskanen
Date: 2024-10-26
"""

from __future__ import annotations
from typing import List


class URL:
    """Hierarchical path for identifying objects in a hierarchy."""

    def __init__(self, path: str = "") -> None:
        """Initializes the URL with an optional starting path.

        Args:
            path (str): The initial path as a string (e.g., "/foo/bar").
        """
        self.segments: List[str] = [
            segment for segment in path.strip("/").split("/") if segment
        ]
        self.is_rooted: bool = path.startswith("/")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, URL):
            return self.segments == other.segments
        return False

    def __str__(self) -> str:
        return f"/{'/'.join(self.segments)}"

    def push_tail(self, name: str) -> None:
        """Adds a segment to the end of the URL path (child).

        Args:
            name (str): The name of the segment to add.
        """
        self.segments.append(name)

    def push_head(self, name: str) -> None:
        """Adds a segment to the beginning of the URL path (parent).

        Args:
            name (str): The name of the segment to add.
        """
        self.segments.insert(0, name)

    def pop_tail(self) -> str:
        """Removes and returns the last segment of the URL path.

        Returns:
            str: The last segment of the path.

        Raises:
            IndexError: If the path is empty.
        """
        if self.segments:
            return self.segments.pop()
        raise IndexError("Cannot pop from an empty URL")

    def pop_head(self) -> str:
        """Removes and returns the first segment of the URL path.

        Returns:
            str: The first segment of the path.

        Raises:
            IndexError: If the path is empty.
        """
        if self.segments:
            return self.segments.pop(0)
        raise IndexError("Cannot pop from an empty URL")

    def get(self) -> str:
        """Gets the full URL path as a string.

        Returns:
            str: The full path, starting with a forward slash if absolute.
        """
        prefix: str = "/" if self.is_rooted else ""
        return prefix + "/".join(self.segments)

    def normalize(self) -> URL:
        """Simplifies the URL path by resolving "." and "..".

        Returns:
            URL: The normalized path.
        """
        stack: List[str] = []
        for segment in self.segments:
            if segment == "..":
                if stack:
                    stack.pop()  # Go up one level
                elif not self.is_rooted:
                    stack.append("..")  # Preserve leading ".." for relative paths
            elif segment != ".":
                stack.append(segment)
        self.segments = stack
        return self

    def is_absolute(self) -> bool:
        """Checks if the URL path is absolute.

        Returns:
            bool: True if the path starts with "/", False otherwise.
        """
        return self.is_rooted

    def make_absolute(self, base: URL) -> URL:
        """Makes the path absolute using a given base path.

        Args:
            base (URL): The base path to combine with.

        Returns:
            URL: The absolute path.
        """
        if self.is_absolute():
            return self
        new_path: URL = base.copy()
        new_path.segments.extend(self.segments)
        return new_path.normalize()

    def starts_with(self, segment: str) -> bool:
        """Checks if the path starts with the given segment.

        Args:
            segment (str): The segment to check.

        Returns:
            bool: True if the path starts with the segment, False otherwise.
        """
        return bool(self.segments and self.segments[0] == segment)

    def prepend_base(self, base: str) -> None:
        """Prepends a base path if the current path is not absolute.

        Args:
            base (str): The base path to prepend.
        """
        if not self.is_absolute():
            base_segments: List[str] = [
                segment for segment in base.strip("/").split("/") if segment
            ]
            self.segments = base_segments + self.segments
            self.is_rooted = base.startswith("/")

    def is_empty(self) -> bool:
        """Checks if the URL path has no segments.

        Returns:
            bool: True if the path is empty, False otherwise.
        """
        return len(self.segments) == 0

    def copy(self) -> URL:
        """Creates a deep copy of the URL.

        Returns:
            URL: A new instance with the same segments.
        """
        return URL(self.get())

    def __repr__(self) -> str:
        """Returns the formal string representation of the URL.

        Returns:
            str: A string representation suitable for debugging.
        """
        return f"URL({self.get()!r})"



"""
Author: Juha Meskanen
Date: 2024-10-26
"""

import argparse
import os
import sys
from typing import List, Type, Optional, Union

from typing_extensions import override

from .format import Format
from .argmaestro import ArgMaestro
from .plugmaster import PlugMaster
from .composite import Composite
from .masterpiece import MasterPiece
from .log import Log
from .treevisualizer import TreeVisualizer


class Application(Composite):
    """Masterpiece application class. Implements startup argument parsing,
    plugin management and initialization of class attributes through
    class specific configuration files.

    """

    plugins: List[Type[MasterPiece]] = []
    serialization_file: str = ""
    serialization_format: str = "JsonFormat"
    plugin_groups = ["masterpiece"]  # plugin group, for Python's module discovery
    color: str = "yellow"
    log_level: int = 0

    _plugmaster: Optional[PlugMaster] = None
    _argmaestro: Optional[ArgMaestro] = None
    _init: bool = False  # initialize class configuration files
    _app_id: str = "masterpiece"  # default application id
    _config = "config"  # default configuration

    def __init__(self, name: str, payload: Optional[MasterPiece] = None) -> None:
        """Instantiates and initializes. By default, the application log
        filename is set to the same as the application name.

        Args:
            name (str): The name of the application, determining the default log filename.
            payload (MasterPiece): Playload object associated with this object.
        """
        super().__init__(name, payload)

    @classmethod
    def get_plugmaster(cls) -> Optional[PlugMaster]:
        """Fetch the plugmaster object reponsible for plugin management.

        Returns:
            PlugMaster: object
        """
        return cls._plugmaster

    @classmethod
    def set_plugmaster(cls, plugmaster: PlugMaster) -> None:
        """Set the plugmaster object reponsible for plugin management.

        Args:
            plugmaster (PlugMaster): object managing plugins
        """
        cls._plugmaster = plugmaster

    @classmethod
    def get_argmaestro(cls) -> Optional[ArgMaestro]:
        """Fetch the plugmaster object reponsible for plugin management.

        Returns:
            PlugMaster: object
        """
        return cls._argmaestro

    @classmethod
    def set_argmaestro(cls, argmaestro: ArgMaestro) -> None:
        """Set the argmaestro object reponsible for plugin management.

        Args:
            argmaestro (PlugMaster): object managing plugins
        """
        cls._argmaestro = argmaestro

    @classmethod
    def get_configuration_filename(cls, name: str) -> str:
        """Generate the user specific file name of the configuration file based on the class name.

        Args:
            name (str): object managing plugins

        """
        return os.path.join(
            os.path.expanduser("~"), "." + cls._app_id, cls._config, name
        )

    @classmethod
    def save_configuration(cls) -> None:
        """Create class configuration file, if configuration is enabled and
        if the file does not exist yet. See --config startup argument.
        """
        filename: str = "undefined"
        if cls.serialization_format == "":
            cls.log_warning(f"No serialization format specified, configuration skipped")
            return
        cls.log_info(f"Saving configuration")

        format_class: Optional[Type[MasterPiece]] = MasterPiece.factory().get(
            cls.serialization_format
        )
        if format_class is not None and issubclass(format_class, Format):
            file_ext: str = format_class.file_extension
            for name, ctor in MasterPiece.factory().items():
                if ctor is not None:
                    try:
                        filename = cls.get_configuration_filename(name) + file_ext

                        # Ensure the target directory exists
                        directory = os.path.dirname(filename)
                        if directory and not os.path.exists(directory):
                            os.makedirs(directory, exist_ok=True)

                        with open(filename, "w", encoding="utf-8") as f:
                            format = format_class(f)
                            format.save_configuration(ctor)
                            cls.log_info(f"Configuration file {filename} saved")
                    except Exception as e:
                        cls.log_error(f"Error in saving {name}:{filename} {e}")

    @classmethod
    def load_configuration(cls) -> None:
        """Load class attributes from a configuration file."""
        filename: str = "undefined"
        if cls.serialization_format == "":
            cls.log_warning(f"No serialization format specified, configuration skipped")
            return
        cls.log_info(f"Loading configuration using {cls.serialization_format}")
        cls.parse_args()
        format_class: Type[MasterPiece] = MasterPiece.factory()[
            cls.serialization_format
        ]
        if format_class is not None and issubclass(format_class, Format):
            file_ext: str = format_class.file_extension
            for name, ctor in MasterPiece.factory().items():
                if ctor is not None:
                    filename = cls.get_configuration_filename(name) + file_ext
                    try:
                        with open(filename, "r", encoding="utf-8") as f:
                            format = format_class(f)
                            format.load_configuration(ctor)
                            cls.log_info(f"Configuration file {filename} loaded")
                    except Exception as e:
                        cls.log_error(f"Error reading {filename}, {e}")

    @classmethod
    def parse_args(cls) -> None:
        """Register classes with ArgMaestro."""
        cls._argmaestro = ArgMaestro()
        for c, clazz in MasterPiece.factory().items():
            if clazz is not None:
                cls._argmaestro.add_class_arguments(clazz)
            else:
                cls.log_error(f"None entry in the class factory for {c}")
        cls._argmaestro.parse_args()

    @classmethod
    def init_app_id(cls, app_id: str = "myapp") -> None:
        """
        Initialize application id. Parses initial startup that depend on application id.
        Must be called before any classes are instanced.

        Arguments:
            -a, --app (str): Application ID.
            -c, --config (str): Configuration name, empty string for no configuration
            -i, --init (bool): Whether to create class configuration files if not already created.
        """
        Application._app_id = app_id
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("-a", "--app", type=str, help="Application ID")
        parser.add_argument("-c", "--config", type=str, help="Configuration")
        parser.add_argument(
            "-l",
            "--log_level",
            type=str,
            help="Log level - DEBUG, INFO,  WARNING, ERROR",
        )
        parser.add_argument(
            "-i",
            "--init",
            action="store_true",
            help="Create class configuration files",
        )
        args, remaining_argv = parser.parse_known_args()
        sys.argv = [sys.argv[0]] + remaining_argv
        MasterPiece.set_log(Log(app_id, Log.parse_level(args.log_level)))
        if args.config:
            Application._config = args.config
        if args.app:
            Application._app_id = args.app
        if args.init:
            Application._init = args.init
            cls.log_info("--init requested, creating class configuration files")
        cls.log_info(f"Configuration files in  ~/.{cls._app_id}/{cls._config}")

        if cls._init:
            cls.log_info(
                "--init specified, class configuration files will be created upon exit"
            )

    @classmethod
    def register_plugin_group(cls, name: str) -> None:
        """Registers a new plugin group within the application. Only plugins that match
        the registered groups will be loaded. By default, all 'masterpiece' plugins
        are included. Frameworks and apps built on the MasterPiece framework can define
        more group names, enabling plugins to be developed for any
        those as well.


        Args:
            name (str): The name of the plugin group to be registered
        """

        if not name in cls.plugin_groups:
            cls.plugin_groups.append(name)

    @classmethod
    def load_plugins(cls) -> None:
        """Loads and initializes all plugins for instantiation. This method
        corresponds to importing Python modules with import clauses."""
        if cls._plugmaster is None:
            cls._plugmaster = PlugMaster(cls.get_app_id())

        for g in cls.plugin_groups:
            cls._plugmaster.load(g)

    def instantiate_plugin_by_name(self, name: str) -> Union[MasterPiece, None]:
        """Installs the plugin by name, that is, instantiates the plugin class
        and inserts the instance as child to the application.
        Args:
            name (str): name of the plugin class
        """
        if self._plugmaster is None:
            return None
        return self._plugmaster.instantiate_class_by_name(self, name)

    def install_plugins(self) -> None:
        """Installs plugins into the application by invoking the `install()` method
        of each loaded plugin module.
        **Note:** This method is intended for testing and debugging purposes only.
        In a typical use case, the application should handle the instantiation of classes and
        manage their attributes as needed.
        """
        if self._plugmaster is None:
            self._plugmaster = PlugMaster(self.get_app_id())
        self._plugmaster.install(self)

    def deserialize(self) -> None:
        """Deserialize instances from the startup file specified by 'serialization_file'
        class attribute, or '--file' startup argument.
        """
        if self.serialization_file != "" and self.serialization_format != "":
            self.info(
                f"Deserializing masterpieces from {self.serialization_file} using {self.serialization_format}"
            )
            format_class: Type[MasterPiece] = MasterPiece.factory()[
                self.serialization_format
            ]
            if issubclass(format_class, Format):
                with open(self.serialization_file, "r", encoding="utf-8") as f:
                    format = format_class(f)
                    format.deserialize(self)
                    self.info(f"File {self.serialization_file} successfully read")
            else:
                raise TypeError(
                    f"{self.serialization_format} is not a subclass of Format"
                )
        else:
            self.warning(
                f"No deserialization this time, --serialization_file not specified"
            )

    def serialize(self) -> None:
        """Serialize application state to the file specified by 'serialization_file'
        class attribute'.
        """
        if self.serialization_file != "":
            self.info(
                f"Saving masterpieces to {self.serialization_file} of type {self.serialization_format}"
            )
            format_class: Type[MasterPiece] = MasterPiece.factory()[
                self.serialization_format
            ]
            if issubclass(format_class, Format):
                with open(self.serialization_file, "w", encoding="utf-8") as f:
                    format = format_class(f)
                    format.serialize(self)
                    self.info(f"File {self.serialization_file} successfully written")
            else:
                raise TypeError(
                    f"{self.serialization_format} is not a subclass of Format"
                )
        else:
            self.warning("No serialization this time, --serialization_file not set")

    @classmethod
    def get_app_id(cls) -> str:
        """Fetch the application id.  Application id determines the folder
        in which the configuration files for classes are held. Note that
        a single package can ship with more than just one executable application,
        all with the same application id.

        ..todo: Application id '_app_id' is prefixed with '_' to signal that it is a
        private attribute (python) and that should not be serialized (masterpiece).
        Isn't there something like @transient in Python? App id needs to be accessed
        outside, which is why this get_app_id() method is needed.

        Returns:
            str: application id determign application registry for class attribute serialization
        """
        return cls._app_id

    @override
    def run(self) -> None:
        if self._init:
            self.save_configuration()
        else:
            super().run()

    @override
    def run_forever(self) -> None:
        if self._init:
            self.save_configuration()
        else:
            super().run_forever()

    def print(self) -> None:
        """
        Print the instance hierarchy of the application using
        `TreeVisualizer`
        """
        visualizer1 = TreeVisualizer(self.color)
        visualizer1.print_tree(self)



"""
Author: Juha Meskanen
Date: 2024-10-26
"""

from typing import Any
import unittest

from io import StringIO
from unittest.mock import patch
from masterpiece.masterpiece import MasterPiece
from masterpiece.composite import Composite
from masterpiece.treevisualizer import TreeVisualizer


class TestComposite(unittest.TestCase):
    """Unit tests for `Composite` class."""

    def setUp(self) -> None:
        """Set up test data for each test."""
        self.composite = Composite(name="group")
        self.child1 = MasterPiece(name="child1")
        self.child2 = MasterPiece(name="child2")

    @patch("sys.stdout", new_callable=StringIO)
    def test_print_with_stdout_capture(self, mock_stdout: Any) -> None:
        """Test the hierarchical print method by capturing stdout."""
        self.composite.add(self.child1)
        self.composite.add(self.child2)
        visualizer = TreeVisualizer("blue")
        visualizer.print_tree(self.composite)

        # Capture the printed output
        output = mock_stdout.getvalue().strip()

        # Define the expected output, taking into account the correct spacing and symbols
        expected_output = (
            "\x1b[34mgroup\x1b[0m\n"
            "    \x1b[34m├─ \x1b[34mchild1\x1b[0m\n"
            "    \x1b[34m└─ \x1b[34mchild2\x1b[0m"
        )
        # Debugging prints
        print("Actual Output:")
        print(repr(output))
        print("Expected Output:")
        print(repr(expected_output))

        # Assert the expected output matches the actual printed output
        self.assertEqual(output, expected_output)


if __name__ == "__main__":
    unittest.main()



"""
Automated class configuration through startup arguments.

Author: Juha Meskanen
Date: 2024-10-26
"""

import argparse
from typing import Type, Union, get_type_hints

from .masterpiece import MasterPiece


class ArgMaestro:
    """
    Automated class configuration through startup arguments.
    ArgMaestro manages startup argument creation and parsing for the class attribute initialization
    of the registered classes. Each startup argument is of form '--[app]_[class]_[attr]'.
    Note: currently supports only int, str, bool and float types.
    """

    def __init__(self) -> None:
        """Construct ArgMaestro with a single argument parser"""
        self.parser = argparse.ArgumentParser(description="Application arguments")
        self.args: dict[str, Type[MasterPiece]] = {}

    def add_class_arguments(self, clazz: Type[MasterPiece]) -> None:
        """Create startup arguments for a class's attributes.

        Args:
            clazz (class): The class to generate arguments for.
        """
        type_hints = get_type_hints(clazz)
        self.args[clazz.__name__] = clazz  # Store the class for future use

        # Filter attributes to ensure only class-specific ones are considered
        for attr, attr_type in type_hints.items():
            try:
                if not attr.startswith("_") and not callable(
                    getattr(clazz, attr, None)
                ):
                    # Skip attributes that are not in the class's own __dict__
                    if attr not in clazz.__dict__:
                        continue

                    # Check attribute type and set default if necessary
                    default_value = getattr(clazz, attr, None)
                    if default_value is None:
                        # Assign logical defaults if None is found
                        if attr_type is float:
                            default_value = 0.0
                        elif attr_type is int:
                            default_value = 0
                        elif attr_type is str:
                            default_value = ""

                    # Argument type handling
                    arg_type: Type[Union[int, float, bool, str]]
                    if attr_type is int:
                        arg_type = int
                    elif attr_type is float:
                        arg_type = float
                    elif attr_type is bool:
                        self.parser.add_argument(
                            f"--{clazz.__name__.lower()}_{attr}",
                            action="store_true",
                            help=f"Enable or disable {attr} in {clazz.__name__}",
                        )
                        print(f"--{clazz.__name__.lower()}_{attr}")
                        continue  # Boolean flag added without needing a default or type
                    else:
                        arg_type = str  # Default to string for unspecified types

                    self.parser.add_argument(
                        f"--{clazz.__name__.lower()}_{attr}",
                        type=arg_type,
                        default=default_value,
                        help=f"Set {attr} in {clazz.__name__} (type: {attr_type.__name__})",
                    )
                    print(
                        f"--{clazz.__name__.lower()}_{attr}, type:{arg_type} default: {default_value}"
                    )
            except Exception as e:
                print(f"Error {e} in parsing argument {attr}")

    def parse_args(self) -> None:
        """Parse arguments and assign values to each class's attributes."""
        # print(self.parser.format_help())
        args = self.parser.parse_args()
        args_dict = vars(args)

        for class_name, clazz in self.args.items():
            # Assign parsed values to class attributes
            for arg, value in args_dict.items():
                prefix = class_name.lower() + "_"
                if arg.startswith(prefix):
                    attr = arg[len(prefix) :]
                    setattr(clazz, attr, value)



import inspect
import json
from typing import Type
import unittest
from io import StringIO

from masterpiece.format import Format
from masterpiece.masterpiece import MasterPiece

from masterpiece.composite import Composite


class TestFormat(unittest.TestCase):

    def setUp(self) ->None:
        """Empty by now."""

    def test_abstractness(self) -> None:
        """Test abstractness of the JsonFormat and its super class."""
        self.assertTrue(inspect.isabstract(Format))
        cl: Type[MasterPiece] = MasterPiece.factory()["JsonFormat"]
        self.assertFalse(inspect.isabstract(cl))


if __name__ == "__main__":
    unittest.main()



import json
from threading import Thread, Event
import threading
import time
from typing import Any, Optional, cast
from typing_extensions import override
from masterpiece import MasterPiece
from masterpiece.mqtt import Mqtt, MqttMsg


class MasterPieceThread(Thread, MasterPiece):
    """Base class for threads used for tasks such as data acquisition
    that need to be run asynchronously. This class defines the `update()`
    method, in which subclasses can execute their specific code. The
    `update_interval()` method (default is 60 seconds) determines how
    frequently the `update()` method is called.

    Args:
        Thread (client): MQTT client for the thread
    """

    systemstatus_topic: str = "system"
    mqtt_root_topic = ""

    event_topic: str = ""

    def __init__(self, client: Optional[Mqtt]) -> None:
        """Construct worker thread for acquiring data and publishing it to
        MQTT

        Args:
            client (Optional[PahoMqtt]): Mqtt client, for communication
        """
        super().__init__()
        MasterPiece.__init__(self)
        self.mqtt_client: Optional[Mqtt] = client
        self.stay = True
        self.name = "unnamed thread"
        self.event_topic = ""
        self._stop_event = Event()
        if self.mqtt_root_topic != "" and self.systemstatus_topic != "":
            self.systemstatus_topic: str = (
                f"{self.mqtt_root_topic}/{self.systemstatus_topic}"
            )
        else:
            self.systemstatus_topic == ""

    def stop(self) -> None:
        """Request the thread to stop processing further tasks.

        Note that the method does not wait the thread to terminate.  If
        the thread is sleeping, it will be awakened and stopped. If the
        thread is in the middle of its code execution, it will finish
        its current job before stopping.  In oder to wait until the
        thread has completed its call join() method.
        """
        self._stop_event.set()

    def run(self) -> None:
        """Thread  loop.

        Calls update() method in a loop and if the return value is True
        sleeps the update_interval() number of seconds before the next
        update call. If the update method returns False then the error
        is logged, and the sleep time is shortened to 5 seconds to
        retry. After three subsequent failures the update_interval is
        reset to original
        """
        self.debug(
            f"Thread {self.name} started with update interval {self.update_interval()}"
        )

        failures: int = 0
        updates: int = 0
        while not self._stop_event.is_set():
            start_time: float = time.time()
            if not self.update():
                seconds: float = 5
                failures = failures + 1
                self.error(
                    f"Thread {self.name} update {str(updates)} failure {str(failures)}, retry after {str(seconds)} ..."
                )
                if failures > 3:
                    failures = 0
                    seconds = self.update_interval()
            else:
                seconds = self.update_interval()
            updates = updates + 1
            self.process_system_status(start_time)
            self._stop_event.wait(seconds)
        self.debug(f"Thread {self.name} stopped")
        # self.mqtt_client = None

    def process_system_status(self, start_time: float) -> None:
        if self.systemstatus_topic != "":
            end_time: float = time.time()
            self.update_metrics(end_time - start_time)
            if self._elapsed > 1.0:
                sysinfo: dict[str, dict[str, float]] = {
                    "threads": {self.name: self.acquire_time_spent()}
                }
                self.publish(
                    self.systemstatus_topic,
                    json.dumps(sysinfo),
                    qos=0,
                    retain=False,
                )

    def update_interval(self) -> float:
        """Fetch the update interval in seconds. The default is 60.

        Returns:
            float: number of seconds
        """
        return 60.0

    def update(self) -> bool:
        """Method called from the threads run loop.

        Up to the sub classes to implement.

        Returns:
            bool: True upon succesfull update. False implies an error .
        """
        return True

    def log(self, type: str, msg: str, details: str) -> None:
        """Log event to event log.

        Args:
            type (str): one of the following: "info", "debug", "warning", "error"
            msg (str): message to be logged
            details (str): detailed description
        """
        if self.mqtt_client is not None:
            data = {"type": type, "msg": msg, "details": details}
            msg = json.dumps(data)
            self.publish(self.event_topic, msg, qos=1, retain=True)

    def publish(
        self, topic: str, message: str, qos: int = 1, retain: bool = True
    ) -> None:
        """Publish the given message to given MQTT topic with specified
        quality of service and retain.

        Args:
            topic (str): topic
            message (str): message to be published
            qos (int): quality of service
            retain (bool): retain the message
        """
        if self.mqtt_client != None:
            mqtt_client: Mqtt = cast(Mqtt, self.mqtt_client)
            mqtt_client.publish(topic, message, qos, retain)

    @override
    def error(self, msg: str, details: str = "") -> None:
        self.log("Error", msg, details)

    @override
    def warning(self, msg: str, details: str = "") -> None:
        self.log("Warning", msg, details)

    @override
    def info(self, msg: str, details: str = "") -> None:
        self.log("Info", msg, details)

    @override
    def debug(self, msg: str, details: str = "") -> None:
        self.log("Debug", msg, details)



"""Example application demonstrating a typical masterpiece application featuring:

- Class initialization through startup arguments 
- Class configuration, with automatically created class configuration files
- Serialization/deserialization of instances
- Visualization of the instance hierarchy

Usage:
------

python myapp.py --init --myapp_serialization_format JsonFormat
python myapp.py 

"""

from masterpiece import Application, MasterPiece, Composite
from typing_extensions import override


class MyApp(Application):
    """Application demonstrating the structure of masterpiece applications.
    Also demonstrates plugin awareness and startup arguments.
    When run, the application prints out its instance hierarchy:

    Example:
        home
        ├─ grid
        ├─ downstairs
        │   └─ kitchen
        │       ├─ oven
        │       └─ fridge
        ├─ garage
        │   └─ EV charger
        └─ solar plant 5.0 kW

    If the --solar [kW] startup argument is passed with a power value, the
    "solar plant" instance is added to the hierarchy.

    """

    solar: float = 0.0

    def __init__(self, name: str = "myapp") -> None:
        """Initialize the home application with the given name.

        Instance attributes can be initialized from class attributes,
        through a serialization file, or from constructor parameters.

        Args:
            name (str): The name of the application.
        """
        super().__init__(name)
        self.create_home()
        self.install_plugins()

    def create_home(self) -> None:
        """Create a default built-in home structure, which can be overridden
        by the instance hierarchy defined in the serialization file. See --file
        startup argument.
        """
        self.create_power_grid()
        self.create_downstairs()
        self.create_garage()
        self.create_solar()

    def create_power_grid(self) -> None:
        """Create the power grid."""
        grid = MasterPiece("grid")
        self.add(grid)

    def create_solar(self) -> None:
        """Create solar plant, if configured by '~/.myapp/config/MyAppApp.json',
        or the '-s' startup argument."""
        if self.solar > 0:
            self.add(MasterPiece(f"solar plant {self.solar} kW"))

    def create_downstairs(self) -> None:
        """Create the downstairs section with a kitchen and appliances."""
        downstairs = Composite("downstairs")
        self.add(downstairs)
        kitchen = Composite("kitchen")
        downstairs.add(kitchen)
        oven = MasterPiece("oven")
        kitchen.add(oven)
        fridge = MasterPiece("fridge")
        kitchen.add(fridge)

    def create_garage(self) -> None:
        """Create the garage with an EV charger."""
        garage = Composite("garage")
        self.add(garage)
        ev_charger = MasterPiece("EV charger")
        garage.add(ev_charger)

    @override
    def run(self) -> None:
        """Start the application."""
        super().run()

        # Print out the instance hierarchy
        self.print()


def main() -> None:
    """Main function that initializes, instantiates, and runs the MyApp application."""

    # Class initialization phase so that they can be instantiated with desired properties
    # Make this app plugin-aware. See the 'masterpiece_plugin' project for a minimal plugin example.
    Application.init_app_id("myapp")
    MyApp.load_plugins()
    Application.load_configuration()

    # Create an instance of MyApp application
    home = MyApp("home")
    home.install_plugins()

    # Initialize from the serialization file if specified
    home.deserialize()

    # Start event processing or the application's main loop
    home.run()

    # Save the application's state to a serialization file (if specified)
    home.serialize()


if __name__ == "__main__":
    main()



"""
Author: Juha Meskanen
Date: 2024-10-26
"""

import logging
import argparse
import sys
from logging.handlers import TimedRotatingFileHandler


class Log(logging.Logger):
    """
    Default `logging.Logger`-based logger implementation for logging events to the MasterPiece
    application log and/or the console.

    TODO: While `logging.Logger` is the de facto logging implementation in Python, many Python components
    have undergone repeated deprecations. As a result, we should not assume that it will remain stable.
    Abstracting the logger API is a simple process that takes just a few minutes. By investing this small
    amount of time, we can shield our application code from potential quirks and deprecations introduced
    by third-party libraries, ensuring greater stability and flexibility in the long run.
    """

    # move these into a abstract base class for loggers
    DEBUG: int = logging.DEBUG
    INFO: int = logging.INFO
    WARNING: int = logging.WARNING
    ERROR: int = logging.ERROR
    CRITICAL: int = logging.CRITICAL

    def __init__(self, name: str, level: int = INFO) -> None:
        """Creates and initializes default logger with the given name and
        logging level. Typically the name is the name of the application.

        Args:
            name: name of the logger
            level: logging level, the default is logging.DEBUG
        """
        super().__init__(name, level)

        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "-f",
            "--log_name",
            type=str,
            default=name + ".log",
            help="Name of the log file",
        )
        parser.add_argument(
            "-l", "--log_level", type=int, default=logging.DEBUG, help="Logging level"
        )
        args, remaining_argv = parser.parse_known_args()
        sys.argv = [sys.argv[0]] + remaining_argv

        self.setLevel(args.log_level)  # Set the logging level here

        # Create file formatter
        file_formatter = logging.Formatter(
            "%(asctime)s [%(levelname)-5.5s]  %(message)-0.280s"
        )

        # Create a timed rotating file handler
        file_handler = TimedRotatingFileHandler(
            args.log_name, when="midnight", interval=1, backupCount=7
        )
        file_handler.setFormatter(file_formatter)
        self.addHandler(file_handler)

        # Create console formatter
        console_formatter = logging.Formatter("[%(levelname)-5.5s]  %(message)-0.280s")
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(console_formatter)
        self.addHandler(console_handler)

    def close(self) -> None:
        """Close all handlers to free resources."""
        for handler in self.handlers:
            handler.close()
            self.removeHandler(handler)

    @classmethod
    def parse_level(cls, level: str) -> int:
        """Map the given symbolic log level to log level value.

        Args:
            level (str): DEBUG, WARNING, INFO, ERROR etc.

        Returns:
            int: Log.DEBUG etc.
        """
        if level == "DEBUG":
            return cls.DEBUG
        elif level == "INFO":
            return cls.INFO
        elif level == "WARNING":
            return cls.WARNING
        elif level == "ERROR":
            return cls.ERROR
        else:
            return cls.INFO



import inspect
import json
from typing import Type
import unittest
from io import StringIO

from masterpiece.format import Format
from masterpiece.masterpiece import MasterPiece
from masterpiece.jsonformat import JsonFormat
from masterpiece.composite import Composite


class TestJsonFormat(unittest.TestCase):

    def setUp(self) ->None:
        """Set up a hierarchical structure for testing."""
        self.parent = Composite("parent")
        self.child = MasterPiece("child")
        self.parent.add(self.child)

    def test_serialize(self) ->None:
        """Test serialization of the hierarchical object to JSON."""
        # Create a StringIO object to simulate a file
        stream = StringIO()
        json_format = JsonFormat(stream)

        # Serialize the parent object
        json_format.serialize(self.parent)

        # Verify the JSON output
        expected_output = json.dumps(self.parent.to_dict(), indent=4)
        self.assertEqual(stream.getvalue().strip(), expected_output)

    def test_deserialize(self) ->None:
        """Test deserialization of JSON back to the hierarchical object."""
        # Create a StringIO object to simulate a file
        stream = StringIO()

        # Serialize the parent object to JSON first
        json_format = JsonFormat(stream)
        json_format.serialize(self.parent)

        # Prepare a new Composite object for deserialization
        new_parent = Composite("")

        # Seek to the beginning of the stream for reading
        stream.seek(0)

        # Deserialize the JSON back to the new parent object
        json_format = JsonFormat(stream)
        json_format.deserialize(new_parent)

        # Verify that the new_parent object has the same attributes
        self.assertEqual(new_parent.name, self.parent.name)
        self.assertEqual(len(new_parent.children), len(self.parent.children))
        self.assertEqual(new_parent.children[0].name, self.child.name)

    def test_abstractness(self) -> None:
        """Test abstractness of the JsonFormat and its super class."""
        self.assertFalse(inspect.isabstract(JsonFormat))
        self.assertTrue(inspect.isabstract(Format))
        cl: Type[MasterPiece] = MasterPiece.factory()["JsonFormat"]
        self.assertFalse(inspect.isabstract(cl))


if __name__ == "__main__":
    unittest.main()
